/*********************************************************
 * The publicly exposed MathQuill API.
 ********************************************************/

var API = {}, Options = P(), optionProcessors = {}, Progenote = P(), EMBEDS = {};

/**
 * Interface Versioning (#459, #495) to allow us to virtually guarantee
 * backcompat. v0.10.x introduces it, so for now, don't completely break the
 * API for people who don't know about it, just complain with console.warn().
 *
 * The methods are shimmed in outro.js so that MQ.MathField.prototype etc can
 * be accessed.
 */
function insistOnInterVer() {
  if (window.console) console.warn(
    'You are using the MathQuill API without specifying an interface version, ' +
    'which will fail in v1.0.0. Easiest fix is to do the following before ' +
    'doing anything else:\n' +
    '\n' +
    '    MathQuill = MathQuill.getInterface(1);\n' +
    '    // now MathQuill.MathField() works like it used to\n' +
    '\n' +
    'See also the "`dev` branch (2014–2015) → v0.10.0 Migration Guide" at\n' +
    '  https://github.com/mathquill/mathquill/wiki/%60dev%60-branch-(2014%E2%80%932015)-%E2%86%92-v0.10.0-Migration-Guide'
  );
}
// globally exported API object
function MathQuill(el) {
  insistOnInterVer();
  return MQ1(el);
};
MathQuill.prototype = Progenote.p;
MathQuill.VERSION = "{VERSION}";
MathQuill.interfaceVersion = function(v) {
  // shim for #459-era interface versioning (ended with #495)
  if (v !== 1) throw 'Only interface version 1 supported. You specified: ' + v;
  insistOnInterVer = function() {
    if (window.console) console.warn(
      'You called MathQuill.interfaceVersion(1); to specify the interface ' +
      'version, which will fail in v1.0.0. You can fix this easily by doing ' +
      'this before doing anything else:\n' +
      '\n' +
      '    MathQuill = MathQuill.getInterface(1);\n' +
      '    // now MathQuill.MathField() works like it used to\n' +
      '\n' +
      'See also the "`dev` branch (2014–2015) → v0.10.0 Migration Guide" at\n' +
      '  https://github.com/mathquill/mathquill/wiki/%60dev%60-branch-(2014%E2%80%932015)-%E2%86%92-v0.10.0-Migration-Guide'
    );
  };
  insistOnInterVer();
  return MathQuill;
};
MathQuill.getInterface = getInterface;

var MIN = getInterface.MIN = 1, MAX = getInterface.MAX = 2;
function getInterface(v) {
  if (!(MIN <= v && v <= MAX)) throw 'Only interface versions between ' +
    MIN + ' and ' + MAX + ' supported. You specified: ' + v;

  /**
   * Function that takes an HTML element and, if it's the root HTML element of a
   * static math or math or text field, returns an API object for it (else, null).
   *
   *   var mathfield = MQ.MathField(mathFieldSpan);
   *   assert(MQ(mathFieldSpan).id === mathfield.id);
   *   assert(MQ(mathFieldSpan).id === MQ(mathFieldSpan).id);
   *
   */
  function MQ(el) {
    if (!el || !el.nodeType) return null; // check that `el` is a HTML element, using the
      // same technique as jQuery: https://github.com/jquery/jquery/blob/679536ee4b7a92ae64a5f58d90e9cc38c001e807/src/core/init.js#L92
    var blockNode = Node.getNodeOfElement($(el).children('.mq-root-block')[0]);
    var ctrlr = blockNode && blockNode.controller;
    return ctrlr ? APIClasses[ctrlr.KIND_OF_MQ](ctrlr) : null;
  };
  var APIClasses = {};

  MQ.L = L;
  MQ.R = R;
  MQ.saneKeyboardEvents = saneKeyboardEvents;

  function config(currentOptions, newOptions) {
    if (newOptions && newOptions.handlers) {
      newOptions.handlers = { fns: newOptions.handlers, APIClasses: APIClasses };
    }
    for (var name in newOptions) if (newOptions.hasOwnProperty(name)) {
      var value = newOptions[name], processor = optionProcessors[name];
      currentOptions[name] = (processor ? processor(value) : value);
    }
  }
  MQ.config = function(opts) { config(Options.p, opts); return this; };
  MQ.registerEmbed = function(name, options) {
    if (!/^[a-z][a-z0-9]*$/i.test(name)) {
      throw 'Embed name must start with letter and be only letters and digits';
    }
    EMBEDS[name] = options;
  };

  var AbstractMathQuill = APIClasses.AbstractMathQuill = P(Progenote, function(_) {
    _.init = function(ctrlr) {
      this.__controller = ctrlr;
      this.__options = ctrlr.options;
      this.id = ctrlr.id;
      this.data = ctrlr.data;
    };
    _.__mathquillify = function(classNames) {
      var ctrlr = this.__controller, root = ctrlr.root, el = ctrlr.container;
      ctrlr.createTextarea();

      var contents = el.addClass(classNames).contents().detach();
      root.jQ = $('<span class="mq-root-block"/>').appendTo(el);
      Node.linkElementByBlockId(root.jQ[0], root.id);
      this.latex(contents.text());

      this.revert = function() {
        return el.empty().unbind('.mathquill')
        .removeClass('mq-editable-field mq-math-mode mq-text-mode')
        .append(contents);
      };
    };
    _.config = function(opts) { config(this.__options, opts); return this; };
    _.el = function() { return this.__controller.container[0]; };
    _.text = function() { return this.__controller.exportText(); };
    _.mathspeak = function() { return this.__controller.exportMathSpeak(); };
    _.latex = function(latex) {
      //NNANH 12/3/2018 - Edit - Bug#8615 - E - Change suggestion with and or xor randseed and "|"
      if(latex === latex + ''){
          latex = latex.replace(/(∧)/g, "\\ and\\ "); //have two side spaces
          latex = latex.replace(/(⊻)/g, "\\ xor\\ "); //convert xor -> different latex and then reset them again
          latex = latex.replace(/(∨)/g, "\\ or\\ "); //have two side spaces
          latex = latex.replace(/(\\left\|)/g, "leftASB"); //have two side spaces
          latex = latex.replace(/(\\right\|)/g, "rightASB"); //have two side spaces
          latex = latex.replace(/(\\RandSeed)/g, "RandSeed");
          latex = latex.replace(/\b(xor)\b/g, "SxrS");
          if(!(latex.indexOf("\\ SxrS\\ ") > -1) && (latex.indexOf(" SxrS ") > 0)){
              latex = latex.replace(/(SxrS)/g, "\\ SxrS\\ "); //have two side spaces
          }
          if(!(latex.indexOf("\\ or\\ ") > -1) && (latex.indexOf(" or ") > 0)){
              latex = latex.replace(/(or)/g, "\\ or\\ "); //have two side spaces
          }
          if(!(latex.indexOf("\\ and\\ ") > -1) && (latex.indexOf(" and ") > 0)){
              latex = latex.replace(/(and)/g, "\\ and\\ "); //have two side spaces
          }
          if(!(latex.indexOf("RandSeed\\ ") > -1)){
              latex = latex.replace(/(RandSeed)/g, "RandSeed\\ "); //have two side spaces
          }
          if(!(latex.indexOf("\\ \|\\ ") > -1)){
              latex = latex.replace(/(\|)/g, "\\ |\\ "); //have two side spaces
          }
          latex = latex.replace(/(leftASB)/g, "\\left|"); //have two side spaces
          latex = latex.replace(/(rightASB)/g, "\\right|"); //have two side spaces
          latex = latex.replace(/\b(SxrS)\b/g, "xor");
          // latex = latex.replace(/\b(or)\b/g, "\\quad or \\quad");
          // latex = latex.replace(/\b(SxrS)\b/g, "\\quad xor \\quad");
          // latex = latex.replace(/\b(and)\b/g, "\\quad and \\quad");
          // latex = latex.replace(/(\|)/g, "\\quad | \\quad");
          //NNANH - S - Add - 4/10/2019 Bug#11287 convert old data rightarrow
          latex = latex.replace(/\\Rightarrow/g, "\\Rightarrow ");
          //NNANH - 4/9/2019 - Start - add - change old data integral Bug#10472
          if (latex.indexOf("\\int_") > -1 && latex.indexOf("}{{{") > -1 && latex.indexOf("}}}") > -1) {
              // adjust substitution to be more specific rmf#930
              var regex = /\\\\int_{([^}]*)}\^{([^}]*)}{{{([^}]*)}d{([^}]*)}}}/gm;
              var subst = '\\\\int_{$1}^{$2}{$3}d{$4}';
              latex = latex.replace(regex, subst);
          }
          //NNANH - 4/9/2019 - End - add - change old data integral Bug#10472
      }
        //NNANH 12/3/2018 - Edit - Bug#8615 - E - Change suggestion with and or xor randseed and "|"
      try {
      if (arguments.length > 0) {
        this.__controller.renderLatexMath(latex);
        if (this.__controller.blurred) this.__controller.cursor.hide().parent.blur();
        return this;
      }
      return this.__controller.exportLatex();
       } catch (error) {
            console.log(error);
       }
       this.__controller.renderLatexMath("");
       if (this.__controller.blurred) this.__controller.cursor.hide().parent.blur();
       return this;
    };
    _.sLatexToSVG= function() {
      return this.__controller.exportSLatexToSVG();
    };
    _.sLatex = function() {
        return this.__controller.exportSLatex();
    };
    _.nativeLatex = function() {
        return this.__controller.exportNativeLatex();
    };
    _.html = function() {
      return this.__controller.root.jQ.html()
        .replace(/ mathquill-(?:command|block)-id="?\d+"?/g, '')
        .replace(/<span class="?mq-cursor( mq-blink)?"?>.?<\/span>/i, '')
        .replace(/ mq-hasCursor|mq-hasCursor ?/, '')
        .replace(/ class=(""|(?= |>))/g, '');
    };
    _.reflow = function() {
      this.__controller.root.postOrder(function (node) { node.reflow(); });
      return this;
    };
    _.ignoreNextMousedown = function(fn) {
        this.__controller.cursor.options.ignoreNextMousedown = fn;
    };
  });
  MQ.prototype = AbstractMathQuill.prototype;

  APIClasses.EditableField = P(AbstractMathQuill, function(_, super_) {
    _.__mathquillify = function() {
      super_.__mathquillify.apply(this, arguments);
      this.__controller.editable = true;
      this.__controller.delegateMouseEvents();
      this.__controller.editablesTextareaEvents();
      return this;
    };
    _.focus = function() {
      this.__controller.textarea.focus();
      this.__controller.scrollHoriz();      if (this.__controller.cursor && this.__controller.cursor[L] !== 0 && this.__controller.cursor[L].checkFunctionName) {
        this.__controller.cursor[L].checkFunctionName(this.__controller.cursor.options);
      }
      return this;
    };
    _.blur = function() { this.__controller.textarea.blur(); return this; };
    _.write = function(latex) {
      //NNANH 2/18/2019 Bug#10409 Add - S - disable event drag drop copy
      if((latex === latex + '') && latex.indexOf("\\type") > -1){
        return this;
      }
      //NNANH 2/18/2019 Bug#10409 Add - S - disable event drag drop copy
      var ctrlr = this.__controller.notify(), cursor = ctrlr.cursor;
      if (ctrlr.options.restrictParameterInSummation  && /\bintegral\b/g.test(latex)
         && (ctrlr.cursor.parent.jQ.closest('.mq-to').length>0||ctrlr.cursor.parent.jQ.closest('.mq-from').length>0)) {
            return this;
      }
      //NNANH 12/21/2018 CR#11812 Add - S - disable event drag drop copy
      var mathquillBlock = cursor.parent.jQ;
      if(mathquillBlock && (mathquillBlock.find(".mq-selection").hasClass("mq-selection") || mathquillBlock.is(".mq-root-block")) && mathquillBlock.find(".mq-recursiveType").hasClass("mq-recursiveType")){
        return this;
      }
      //NNANH 12/21/2018 CR#11812 Add - S - disable event drag drop copy
      // if (ctrlr.cursor.parent && ctrlr.cursor.parent.parent &&
      //     ctrlr.cursor.parent.parent.ctrlSeq === "parametric" &&
      //     ctrlr.cursor.parent.jQ.hasClass("accept-only-digit")
      //     && ctrlr.options.restrictInputInParametric) {
      //       var ENABLE_KEY_PARAMETRIC = /[0-9.-]/;
      //       if(!ENABLE_KEY_PARAMETRIC.test(latex)){
      //         return this;
      //       }
      // }
      if (ctrlr.options.restrictInputAfterParametric &&
        (/\bparametric\b/g.test(ctrlr.cursor[L].ctrlSeq) || /\bparametric\b/g.test(ctrlr.cursor[R].ctrlSeq))) {
        return this;
      }
      if (ctrlr.options.restrictInputOnSubOfDiff && /\bdiff\b/g.test(cursor.parent.parent.ctrlSeq)
          && cursor.parent.jQ.parent().is(".mq-sup") && !/^\d+$/g.test(latex)) {
          return this;
      }
      try {
      this.__controller.writeLatex(latex);
      } catch (error) {
        console.log(error);
      }
      this.__controller.scrollHoriz();
      if (this.__controller.blurred) this.__controller.cursor.hide().parent.blur();
      return this;
    };
    _.cmd = function(cmd) {
      var ctrlr = this.__controller.notify(), cursor = ctrlr.cursor;

      if (ctrlr.options.restrictParameterInSummation && /\bintegral\b/g.test(cmd)
         && (ctrlr.cursor.parent.jQ.closest('.mq-to').length>0||ctrlr.cursor.parent.jQ.closest('.mq-from').length>0)) {
            return this;
      }
      if (ctrlr.cursor.parent && ctrlr.cursor.parent.parent &&
          ctrlr.cursor.parent.parent.ctrlSeq === "parametric" &&
          ctrlr.cursor.parent.jQ.hasClass("accept-only-digit")
          && ctrlr.options.restrictInputInParametric) {
            //NNANH Start Bug#6401 enable the number for parametric
            var ENABLE_KEY_PARAMETRIC = /[0-9-.]/;
            if(!ENABLE_KEY_PARAMETRIC.test(cmd)){
            //NNANH Start Bug#6401 enable the number for parametric
              return this;
            }
      }
      if (ctrlr.cursor.parent && ctrlr.cursor.parent.parent && 
          ctrlr.cursor.parent.parent.ctrlSeq === "\\mixedFraction" &&
          ctrlr.cursor.parent.jQ.hasClass("mq-mixedFraction")){
            var ENABLE_KEY_MIXEDFRACTION = /[0-9]/;
            if(!ENABLE_KEY_MIXEDFRACTION.test(cmd)){
              return this;
            }
          }      
      // NNANH 12/21/2018 ADD - S - Disable edit or delete mq-recursivetype
      var mathquillBlock = ctrlr.cursor.parent.jQ;
      if(mathquillBlock && (mathquillBlock.find(".mq-selection").hasClass("mq-selection") || mathquillBlock.is(".mq-root-block")) && mathquillBlock.find(".mq-recursiveType").hasClass("mq-recursiveType")){
        return this;
      }
      // NNANH 12/21/2018 ADD - E - Disable edit or delete mq-recursivetype
      if (ctrlr.options.restrictInputAfterParametric &&
        (/\bparametric\b/g.test(ctrlr.cursor[L].ctrlSeq) || /\bparametric\b/g.test(ctrlr.cursor[R].ctrlSeq))) {
        return this;
      }
      if (ctrlr.options.restrictInputOnSubOfDiff && /\bdiff\b/g.test(cursor.parent.parent.ctrlSeq)
          && cursor.parent.jQ.parent().is(".mq-sup") && !/^\d+$/g.test(cmd)) {
          return this;
      }
      //// add for financial start <murayama> 20190611.
      //if ( ctrlr.cursor.parent &&
      //     ctrlr.cursor.parent.jQ.hasClass("accept-only-digit") ) {
      //      var ENABLE_KEY_PARAMETRIC = /[0-9-.]/;
      //      if(!ENABLE_KEY_PARAMETRIC.test(cmd)){
      //        return this;
      //      }
      //}
      //// add for financial end   <murayama> 20190611.
      // add for financial start <murayama> 20190704.
      if ( ctrlr.cursor.parent &&
           ctrlr.cursor.parent.jQ.hasClass("accept-read-only") ) {
          console.log("accept-read-only: "+cmd);
          return this;
      }
      // add for financial end   <murayama> 20190704.
      //for command
      if (/^\\[a-z]+$/i.test(cmd)) {
        if (cursor.parent.parent.sub && !ctrlr.options.autoBreakOutOfSubscript
           && cursor.parent.jQ.is(".mq-sub")
           && ctrlr.options.restrictInputOfSubscript) {
            return this;
        }
        if (cursor.parent.parent.sub && cursor.parent.jQ.is(".mq-sub") && ctrlr.options.autoBreakOutOfSubscript) {
          var currentParent = cursor.parent.parent;
          while (currentParent.sub) {
            if (currentParent.parent.parent.sub) {
              currentParent = currentParent.parent.parent;
            } else {
              currentParent = currentParent.parent;
            }
          }
          cursor.insAtRightEnd(currentParent);
        }
        cmd = cmd.slice(1);
        var klass = LatexCmds[cmd] || Environments[cmd];
        if (klass) {
          cmd = klass(cmd);
          if (cursor.selection) cmd.replaces(cursor.replaceSelection());
          cmd.createLeftOf(cursor.show());
        }
        else /* TODO: API needs better error reporting */;
      }
      else cursor.parent.write(cursor, cmd);

      ctrlr.scrollHoriz();
      if (ctrlr.blurred) cursor.hide().parent.blur();
      return this;
    };
    _.select = function() {
      var ctrlr = this.__controller;
      ctrlr.notify('move').cursor.insAtRightEnd(ctrlr.root);
      while (ctrlr.cursor[L]) ctrlr.selectLeft();
      return this;
    };
    _.clearSelection = function() {
      this.__controller.cursor.clearSelection();
      return this;
    };

    _.moveToDirEnd = function(dir) {
      this.__controller.notify('move').cursor.insAtDirEnd(dir, this.__controller.root);
      return this;
    };
    _.moveToLeftEnd = function() { return this.moveToDirEnd(L); };
    _.moveToRightEnd = function() { return this.moveToDirEnd(R); };

    _.keystroke = function(keys, evt) {
      var keys = keys.replace(/^\s+|\s+$/g, '').split(/\s+/);
      for (var i = 0; i < keys.length; i += 1) {
        this.__controller.keystroke(keys[i], evt || { preventDefault: noop });
      }
      return this;
    };
    _.typedText = function(text) {
      var ctrlr = this.__controller.notify(), cursor = ctrlr.cursor;
      // if (ctrlr.cursor.parent && ctrlr.cursor.parent.parent &&
      //     ctrlr.cursor.parent.parent.ctrlSeq === "parametric" &&
      //     ctrlr.cursor.parent.jQ.hasClass("accept-only-digit")
      //     && ctrlr.options.restrictInputInParametric) {
      //       var ENABLE_KEY_PARAMETRIC = /[0-9.-]/;
      //       if(!ENABLE_KEY_PARAMETRIC.test(text)){
      //         return this;
      //       }
      // }
      if (ctrlr.options.restrictInputAfterParametric &&
        (/\bparametric\b/g.test(ctrlr.cursor[L].ctrlSeq) || /\bparametric\b/g.test(ctrlr.cursor[R].ctrlSeq))) {
        return this;
      }
      if (ctrlr.options.restrictInputOnSubOfDiff && /\bdiff\b/g.test(cursor.parent.parent.ctrlSeq)
         && cursor.parent.jQ.parent().is(".mq-sup") && !/^\d+$/g.test(text)) {
        return this;
      }
      for (var i = 0; i < text.length; i += 1) this.__controller.typedText(text.charAt(i));
      return this;
    };
    _.dropEmbedded = function(pageX, pageY, options) {
      var clientX = pageX - $(window).scrollLeft();
      var clientY = pageY - $(window).scrollTop();

      var el = document.elementFromPoint(clientX, clientY);
      this.__controller.seek($(el), pageX, pageY);
      var cmd = Embed().setOptions(options);
      cmd.createLeftOf(this.__controller.cursor);
    };
    _.setAriaLabel = function(ariaLabel) {
      if(ariaLabel && typeof ariaLabel === 'string' && ariaLabel!='') this.__controller.ariaLabel = ariaLabel;
      else this.__controller.ariaLabel = 'Math Input';
      return this;
    };
    _.getAriaLabel = function () {
      return this.__controller.ariaLabel || 'Math Input';
    };
    _.setAriaPostLabel = function(ariaPostLabel, timeout) {
      var controller = this.__controller;
      if(ariaPostLabel && typeof ariaPostLabel === 'string' && ariaPostLabel!='') {
        if (
          ariaPostLabel !== controller.ariaPostLabel &&
          typeof timeout === 'number'
        ) {
          if (this.ariaAlertTimeout) clearTimeout(this.ariaAlertTimeout);
          this.ariaAlertTimeout = setTimeout(function() {
            if (!!$(document.activeElement).closest($(controller.container)).length) {
              aria.alert(this.mathspeak().trim() + ' ' + ariaPostLabel.trim());
            }
          }.bind(this), timeout);
        }
        controller.ariaPostLabel = ariaPostLabel;
      } else {
        if (this.ariaAlertTimeout) clearTimeout(this.ariaAlertTimeout);
        controller.ariaPostLabel = '';
      }
      return this;
    };
    _.getAriaPostLabel = function () {
      return this.__controller.ariaPostLabel || '';
    };
    _.clickAt = function(clientX, clientY, target) {
      target = target || document.elementFromPoint(clientX, clientY);
      var ctrlr = this.__controller, root = ctrlr.root;
      if (!jQuery.contains(root.jQ[0], target)) target = root.jQ[0];
      ctrlr.seek($(target), clientX + pageXOffset, clientY + pageYOffset);
      if (ctrlr.blurred) this.focus();
      return this;
    };
    _.ignoreNextMousedown = function(fn) {
      this.__controller.cursor.options.ignoreNextMousedown = fn;
      return this;
    };
    _.deleteNode = function(nodeId) {
      var ctrlr = this.__controller.notify(), cursor = ctrlr.cursor;
      var node = Node.byId[nodeId];
      if (node) {
        ctrlr.notify('move').cursor.insAtDirEnd(L, ctrlr.root);
        node.remove();
        delete cursor[R];
        this.__controller.scrollHoriz();
        cursor.parent.bubble(function (node) { node.reflow(); });
      }
      return this;
    };
    //20180608-ltmthuy-add-Start: Task #6295 グラフ式をEdit中、文字が入力されてない部分をクリックしてもカーソルが末尾に移動しない
    _.moveToEnd = function() {
      var ctrlr = this.__controller, root = ctrlr.root, cursor = ctrlr.cursor;
      if (root.ends[1] && root.ends[1] !== 0) {
        cursor.insRightOf(root.ends[1]);
      }
      return this;
    };
  });
  MQ.EditableField = function() { throw "wtf don't call me, I'm 'abstract'"; };
  MQ.EditableField.prototype = APIClasses.EditableField.prototype;

  /**
   * Export the API functions that MathQuill-ify an HTML element into API objects
   * of each class. If the element had already been MathQuill-ified but into a
   * different kind (or it's not an HTML element), return null.
   */
  for (var kind in API) (function(kind, defAPIClass) {
    var APIClass = APIClasses[kind] = defAPIClass(APIClasses);
    MQ[kind] = function(el, opts) {
      var mq = MQ(el);
      if (mq instanceof APIClass || !el || !el.nodeType) return mq;
      var ctrlr = Controller(APIClass.RootBlock(), $(el), Options());
      ctrlr.KIND_OF_MQ = kind;
      return APIClass(ctrlr).__mathquillify(opts, v);
    };
    MQ[kind].prototype = APIClass.prototype;
  }(kind, API[kind]));

  return MQ;
}

MathQuill.noConflict = function() {
  window.MathQuill = origMathQuill;
  return MathQuill;
};
var origMathQuill = window.MathQuill;
window.MathQuill = MathQuill;

function RootBlockMixin(_) {
  var names = 'moveOutOf deleteOutOf selectOutOf upOutOf downOutOf'.split(' ');
  for (var i = 0; i < names.length; i += 1) (function(name) {
    _[name] = function(dir) { this.controller.handle(name, dir); };
  }(names[i]));
  _.reflow = function() {
    this.controller.handle('reflow');
    this.controller.handle('edited');
    this.controller.handle('edit');
  };
  _.moveOutOfBlock = function() {
    this.controller.handle('moveOutOfBlock');
  };
}
